@@ -34,7 +34,8 @@
 
     \section posixdesc DESCRIPTION
 
-The Verify Proxy module will verify the validity of the proxy (with its chain) or any other X.509 certificate chain that has been used in the LCMAPS framework.
+The Verify Proxy module will verify the validity of the proxy (with its chain)
+or any other X.509 certificate chain that has been used in the LCMAPS framework.
 The module will report to the LCMAPS framework the validity of the chain.
 
 */
@@ -51,54 +52,74 @@ The module will report to the LCMAPS framework the validity of the chain.
                             Include header files
 ******************************************************************************/
 
-#include <string.h>
-#include <strings.h>
-#include <stdio.h>
-#include <stdlib.h>
-#include <ctype.h>
-#include <pwd.h>
-#include <time.h>
-#include <sys/stat.h>
+/* Needed for (un)setenv and strdup (for latter 500 is sufficient) */
+#define	_XOPEN_SOURCE	600
 
 #include "lcmaps_verify_proxy_config.h"
+
+#include <string.h>
+#include <strings.h>
+#include <stdlib.h>
+#include <sys/stat.h>
+#include <errno.h>
+#include <time.h>
+
+#include <openssl/x509.h>
+#include <openssl/err.h>
+
 #include <lcmaps/lcmaps_modules.h>
 #include <lcmaps/lcmaps_arguments.h>
-#include <lcmaps/lcmaps_cred_data.h>
-#include <lcmaps/lcmaps_vo_data.h>
-#include <lcmaps/lcmaps_types.h>
 
-/* Verifification lib */
+
+#if defined(HAVE_LCMAPS_LCMAPS_PLUGIN_PROTOTYPES_H)
+#   include <lcmaps/lcmaps_plugin_prototypes.h>
+#else
+#   include "lcmaps_plugin_prototypes.h"
+#endif
+
+/* Verification lib */
 #include "verify_x509.h"
-#include "proxylifetime.h"
 
+/* proxylifetime routines */
+#include "lcmaps_proxylifetime.h"
 
+/* Define value of environment variable name for allowing discarding of the
+ * private key. See man page under --discard_private_key_absence and
+ * --never_discard_private_key_absence */
+#define DISCARD_PRIVATE_KEY_ABSENCE "VERIFY_PROXY_DISCARD_PRIVATE_KEY_ABSENCE"
 
 #define PLUGIN_RUN      0
 #define PLUGIN_VERIFY   1
 
+/* Define default CA certificate directory */
+#ifndef DEFAULT_CERT_DIR
+# define DEFAULT_CERT_DIR   "/etc/grid-security/certificates"
+#endif
+
+/******************************************************************************
+                       Internal prototype
+******************************************************************************/
+/**
+ * actual plugin_run or plugin_verify function, which one depends on last
+ * argument
+ */
+static int plugin_run_or_verify(int, lcmaps_argument_t *, int);
+
 
 /******************************************************************************
                        Define module specific variables
 ******************************************************************************/
 
-int plugin_initialize (int, char **);
-int plugin_introspect(int *, lcmaps_argument_t **);
-int plugin_run(int, lcmaps_argument_t *);
-int plugin_verify(int, lcmaps_argument_t *);
-int plugin_terminate(void);
+/* Used for checking the maximum TTLs */
+static struct TProxyLevelTTL * plt = NULL;  /* max proxy TTL values	    */
+static struct TProxyLevelTTL * vplt = NULL; /* max voms AC TTL values   */
 
-static int plugin_run_or_verify(int, lcmaps_argument_t *, int);
-
-
-
-static struct TProxyLevelTTL * plt = NULL;
-static struct TProxyLevelTTL * vplt = NULL;
-
-static int    allow_limited = 1;
-static int    require_limited = 0;
-static int    only_enforce_lifetime_checks = 0;
-static int    discard_private_key_absence = 0;
-static char * certdir = NULL;
+/* commandline options and their defaults */
+static int    allow_limited = 1;		/* --allow-limited-proxy */
+static int    require_limited = 0;		/* --require-limited-proxy */
+static int    only_enforce_lifetime_checks = 0;	/* --only-enforce-lifetime-checks */
+static int    discard_private_key_absence = 0;  /* --discard_private_key_absence */
+static char * certdir = NULL;			/* -cadir or -certdir */
 
 
 /******************************************************************************
@@ -118,7 +139,7 @@ int plugin_initialize(
         char ** argv
 )
 {
-    char * logstr = "lcmaps_plugin_verify_proxy-plugin_initialize()";
+    const char * logstr = "lcmaps_plugin_verify_proxy-plugin_initialize()";
     int i = 0;
     struct stat s;
     int never_discard_private_key_absence = 0;
@@ -138,10 +159,10 @@ int plugin_initialize(
 
 
     /*
-     * Check if a processes has set the option to allow to discard the private key verification
-     * The environment variable that provided this choice will be cleared
+     * Check if a processes has set the option to allow to discard the private
+     * key verification. The environment variable that provided this choice will
+     * be cleared
      */
-    #define DISCARD_PRIVATE_KEY_ABSENCE "VERIFY_PROXY_DISCARD_PRIVATE_KEY_ABSENCE"
 
     if (getenv (DISCARD_PRIVATE_KEY_ABSENCE))
     {
@@ -156,7 +177,10 @@ int plugin_initialize(
      */
     for (i = 1; i < argc; i++)
     {
-        if ( ((strncasecmp(argv[i], "-cadir", 6) == 0) || (strncasecmp(argv[i], "-certdir", 8) == 0))
+        if ( ((strncasecmp(argv[i], "-cadir", (size_t)6) == 0) ||
+	      (strncasecmp(argv[i], "-certdir", (size_t)8) == 0) ||
+	      (strncasecmp(argv[i], "-capath", (size_t)7) == 0) ||
+	      (strncasecmp(argv[i], "--capath", (size_t)8) == 0))
              && (i + 1 < argc)
            )
         {
@@ -172,34 +196,36 @@ int plugin_initialize(
             i++;
         }
 
-        else if (strncasecmp(argv[i], "--discard_private_key_absence", 30) == 0)
+        else if (strncasecmp(argv[i], "--discard_private_key_absence", (size_t)30) == 0)
         {
             discard_private_key_absence = 1;
         }
 
         /*
-         * The variable "--never_discard_private_key_absence"
-         * will mute the environment variable that can override the private key verification functionality.
-         * The environment variable that would allow for the discard of the check for the private key will
-         * be useless.
-         * This is used in situation where the private key check is mandatory and non-overrideable.
+	 * The variable "--never_discard_private_key_absence" will mute the
+	 * environment variable that can override the private key verification
+	 * functionality.
+	 * The environment variable that would allow for the discard of the
+	 * check for the private key will be useless.
+	 * This is used in situation where the private key check is mandatory
+	 * and non-overrideable.
          */
-        else if (strncasecmp(argv[i], "--never_discard_private_key_absence", 36) == 0)
+        else if (strncasecmp(argv[i], "--never_discard_private_key_absence", (size_t)36) == 0)
         {
             never_discard_private_key_absence = 1;
         }
 
-        else if (strncasecmp(argv[i], "--allow-limited-proxy", 21) == 0)
+        else if (strncasecmp(argv[i], "--allow-limited-proxy", (size_t)21) == 0)
         {
             allow_limited = 1;
         }
 
-        else if (strncasecmp(argv[i], "--disallow-limited-proxy", 24) == 0)
+        else if (strncasecmp(argv[i], "--disallow-limited-proxy", (size_t)24) == 0)
         {
             allow_limited = 0;
         }
 
-        else if (strncasecmp(argv[i], "--require-limited-proxy", 23) == 0)
+        else if (strncasecmp(argv[i], "--require-limited-proxy", (size_t)23) == 0)
         {
             require_limited = 1;
         }
@@ -209,7 +235,7 @@ int plugin_initialize(
             only_enforce_lifetime_checks = 1;
         }
 
-        else if (  (strncasecmp(argv[i], "--max-voms-ttl", 14) == 0) &&
+        else if (  (strncasecmp(argv[i], "--max-voms-ttl", (size_t)14) == 0) &&
                    (i + 1 < argc) )
 
         {
@@ -221,14 +247,14 @@ int plugin_initialize(
             {
                 /* Proxy Time To Life check is activated */
 
-                if ((t_for_time = ttl_char2time_t (argv[i + 1])) <= 0)   /* 2d-13:37 */
+                if ((t_for_time = lcmaps_lifetime_ttl_char2time_t (argv[i + 1])) <= 0)   /* 2d-13:37 */
                 {
                     lcmaps_log(LOG_ERR,"%s: Parse error in initialization parameter: %s. Use format: 2d-13:37\n", logstr, argv[i]);
                     return LCMAPS_MOD_FAIL;
                 }
 
-                lcmaps_log_debug (5, "%s: Proxy TTL set at: %d seconds at proxy level: %d\n", logstr, t_for_time, level);
-                Push_New_Entry (&vplt, level, t_for_time);
+                lcmaps_log_debug (5, "%s: VOMS TTL set at: %ld seconds at proxy level: %d\n", logstr, (long)t_for_time, level);
+                lcmaps_lifetime_Push_New_Entry (&vplt, level, t_for_time);
             }
             else
             {
@@ -237,45 +263,51 @@ int plugin_initialize(
             }
             i++;
         }
-        else if ( ((strncasecmp(argv[i], "--max-proxy-level-ttl=", 22) == 0) && (strlen(argv[i]) == 23) && (i + 1 < argc)) ||
-                  ((strncasecmp(argv[i], "--max-proxy-level-ttl@", 22) == 0) && (strlen(argv[i]) == 23) && (i + 1 < argc)) )
+        else if ( ((strncasecmp(argv[i], "--max-proxy-level-ttl=", (size_t)22) == 0) && (i + 1 < argc)) ||
+                  ((strncasecmp(argv[i], "--max-proxy-level-ttl@", (size_t)22) == 0) && (i + 1 < argc)) )
         {
             int    level = 0;
             time_t t_for_time = 0;
+	    long   longval;
 
 
-            if ((argv[i][strlen(argv[i]) - 1] == 'l') || (argv[i][strlen(argv[i]) - 1] == 'L'))
+            if ( (strlen(argv[i]) == 23) &&
+		 ( (argv[i][22] == 'l') || (argv[i][22] == 'L') ) )
                 level = LEAF_PROXY;
-            else
-                level = atoi(&(argv[i][strlen(argv[i]) - 1]));
+            else    {
+		errno=0;
+		longval=strtol(&(argv[i][22]),NULL,10);
+		if (errno!=0 || longval<0 || longval>=LEAF_PROXY)  {
+		    lcmaps_log(LOG_ERR,"%s: Parse error in initialization parameter: %s.\n", logstr, argv[i]);
+		    return LCMAPS_MOD_FAIL;
+		}
+		level=(int)longval;
+	    }
 
+	    if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
+	    {
+		/* Proxy Time To Life check is activated */
+		if ((t_for_time = lcmaps_lifetime_ttl_char2time_t (argv[i + 1])) <= 0)   /* 2d-13:37 */
+		{
+		    lcmaps_log(LOG_ERR,"%s: Parse error in initialization parameter: %s. Use format: 2d-13:37\n", logstr, argv[i]);
+		    return LCMAPS_MOD_FAIL;
+		}
+		if (level==LEAF_PROXY)
+		    lcmaps_log_debug (5,
+			"%s: Proxy TTL set at: %ld seconds at LEAF proxy level\n",
+			logstr, (long)t_for_time);
+		else
+		    lcmaps_log_debug (5,
+			"%s: Proxy TTL set at: %ld seconds at proxy level: %d\n",
+			logstr, (long)t_for_time, level);
 
-            if ( (level >= 0) || (level <= 9) || (level == LEAF_PROXY))
-            {
-                if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
-                {
-                    // Proxy Time To Life check is activated
-                    lcmaps_log_debug (5, "%s: Proxy TTL set at: %d seconds at proxy level: %d\n", logstr, t_for_time, level);
-
-                    if ((t_for_time = ttl_char2time_t (argv[i + 1])) <= 0)   // 2d-13:37
-                    {
-                        lcmaps_log(LOG_ERR,"%s: Parse error in initialization parameter: %s. Use format: 2d-13:37\n", logstr, argv[i]);
-                        return LCMAPS_MOD_FAIL;
-                    }
-
-                    Push_New_Entry (&plt, level, t_for_time);
-                }
-                else
-                {
-                    lcmaps_log(LOG_ERR,"%s: Parse error in initialization parameter: %s. It's NULL string or something undefined.\n", logstr, argv[i + 1]);
-                    return LCMAPS_MOD_FAIL;
-                }
-            }
-            else
-            {
-                lcmaps_log(LOG_ERR,"%s: Parse error in initialization parameter: %s. Last character is not a single digit number.\n", logstr, argv[i]);
-                return LCMAPS_MOD_FAIL;
-            }
+		lcmaps_lifetime_Push_New_Entry (&plt, level, t_for_time);
+	    }
+	    else
+	    {
+		lcmaps_log(LOG_ERR,"%s: Parse error in initialization parameter: %s. It's NULL string or something undefined.\n", logstr, argv[i + 1]);
+		return LCMAPS_MOD_FAIL;
+	    }
 
             i++;
         }
@@ -287,6 +319,19 @@ int plugin_initialize(
         }
     }
 
+    /* Set certdir from other means if not specified on the cmdline
+     * including gLExec->LCMAPS */
+    if (certdir==NULL)	{
+	char *value=getenv("X509_CERT_DIR");
+	certdir=value ? strdup(value) : strdup(DEFAULT_CERT_DIR);
+	if (certdir==NULL)  {
+	    lcmaps_log(LOG_ERR,"%s: Out of memory\n",logstr);
+	    return LCMAPS_MOD_FAIL;
+        }
+	lcmaps_log(LOG_INFO,"%s: Using certificate CA Path: %s\n",
+		logstr, certdir); 
+    }
+
     /* Finalization check */
 
     /* Disallowed and required, equals XOR config */
@@ -297,8 +342,8 @@ int plugin_initialize(
     }
 
 
-    Print_TTL_By_Level (plt);
-    Print_TTL_By_Level (vplt);
+    lcmaps_lifetime_Print_TTL_By_Level (plt);
+    lcmaps_lifetime_Print_TTL_By_Level (vplt);
 
 
     /*
@@ -330,7 +375,7 @@ int plugin_introspect(
         lcmaps_argument_t ** argv
 )
 {
-    char * logstr = "lcmaps_plugin_verify-proxy-plugin_introspect()";
+    const char * logstr = "lcmaps_plugin_verify-proxy-plugin_introspect()";
     static lcmaps_argument_t argList[] = {
         { "px509_cred"     , "X509 *"                ,  0, NULL},
         { "px509_chain"    , "STACK_OF(X509) *"      ,  0, NULL},
@@ -345,7 +390,7 @@ int plugin_introspect(
     *argv = argList;
     lcmaps_log_debug(5,"%s: before lcmaps_cntArgs()\n", logstr);
     *argc = lcmaps_cntArgs(argList);
-    lcmaps_log_debug(5,"%s: address first argument: 0x%x\n", logstr, argList);
+    lcmaps_log_debug(5,"%s: address first argument: %p\n", logstr, (void*)argList);
 
     lcmaps_log_debug(3,"%s: Introspect succeeded\n", logstr);
     return LCMAPS_MOD_SUCCESS;
@@ -399,23 +444,51 @@ int plugin_verify(
 }
 
 
+/******************************************************************************
+Function:   plugin_terminate
+Description:
+    Terminate plugin
+Parameters:
+
+Returns:
+    LCMAPS_MOD_SUCCESS : succes
+    LCMAPS_MOD_FAIL    : failure
+******************************************************************************/
+int plugin_terminate()
+{
+    const char * logstr = "lcmaps_plugin_verify_proxy-plugin_terminate()";
+    lcmaps_log_debug(4,"%s: terminating\n", logstr);
+
+    /* if (vomsdir) free(vomsdir); */
+    if (certdir) free(certdir);
+
+    lcmaps_lifetime_FreeProxyLevelTTL (&plt);
+    lcmaps_lifetime_FreeProxyLevelTTL (&vplt);
+
+    return LCMAPS_MOD_SUCCESS;
+}
+
+
+/******************************************************************************
+                       Internal function(s)
+******************************************************************************/
+
 static int plugin_run_or_verify(
         int argc,
         lcmaps_argument_t * argv,
         int lcmaps_mode
 )
 {
-    char             *  logstr      = "lcmaps_plugin_verify_proxy-plugin_run()";
+    const char       *  logstr      = "lcmaps_plugin_verify_proxy-plugin_run()";
 
-    time_t              myclock;
-    struct tm        *  tmpTime;
 
     STACK_OF(X509)   *  myChain = NULL;
     X509             *  lastCert = NULL;
-    lcmaps_proxy_type_t certificate_type = NONE;
-    int                 depth = 0;
+    proxy_type_t        certificate_type = NONE;
+    int                 depth;
 
-    unsigned long       result = 0;
+    verify_x509_error_t	setparam_result;
+    unsigned long       result,errcode;
 
     char              * pem_string = NULL;
     lcmaps_vomsdata_t * vomsdata  = NULL;
@@ -423,6 +496,7 @@ static int plugin_run_or_verify(
 
     /* Verification data */
     internal_verify_x509_data_t * verify_data;
+    const char       *reason=NULL;
 
 
     /*
@@ -487,8 +561,7 @@ static int plugin_run_or_verify(
         else
         {
             lcmaps_log_debug (5, "%s: vomsdata->nvoms = %d\n", logstr, vomsdata->nvoms);
-            lcmaps_log_debug (5, "%s: vomsdata->voms = %x\n", logstr, vomsdata->voms);
-            lcmaps_log_debug (5, "%s: vomsdata->voms[0] = %x\n", logstr, vomsdata->voms[0]);
+            lcmaps_log_debug (5, "%s: vomsdata->voms = %p\n", logstr, (void*)vomsdata->voms);
             lcmaps_log_debug (5, "%s: vomsdata->voms[0].userdn = %s\n", logstr, vomsdata->voms[0].user_dn);
         }
     }
@@ -516,22 +589,22 @@ static int plugin_run_or_verify(
          * Note: must implement check that make use of the CA namespace files
          */
 
-        result = verify_X509_init(&verify_data);
-        set_log_level (5);
+	if (verify_X509_init(&verify_data))
+            goto fail_verify_proxy;
 
-        result = verify_X509_setParameter (&verify_data, VERIFY_X509_STACK_OF_X509, myChain);
-        if (result == ERR_VERIFY_X509_PARAMS_ALREADY_SET)
-            Log (L_WARN, "VERIFY_X509_CERTIFICATE_FILEPATH already set...\n");
-        else if (result == ERR_VERIFY_X509_PARAMS_ACCESS_FAILURE)
+        setparam_result = verify_X509_setParameter (&verify_data, VERIFY_X509_STACK_OF_X509, myChain);
+        if (setparam_result == VER_R_X509_PARAMS_ALREADY_SET)
+            lcmaps_log (LOG_WARNING, "%s: VERIFY_X509_CERTIFICATE_FILEPATH already set...\n",logstr);
+        else if (setparam_result != VER_R_X509_PARAMS_OK)
         {
             verify_X509_term(&verify_data);
             goto fail_verify_proxy;
         }
 
-        result = verify_X509_setParameter (&verify_data, VERIFY_X509_CA_PATH, certdir);
-        if (result == ERR_VERIFY_X509_PARAMS_ALREADY_SET)
-            Log (L_WARN, "VERIFY_X509_CA_PATH already set...\n");
-        else if (result == ERR_VERIFY_X509_PARAMS_ACCESS_FAILURE)
+        setparam_result = verify_X509_setParameter (&verify_data, VERIFY_X509_CA_PATH, certdir);
+        if (setparam_result == VER_R_X509_PARAMS_ALREADY_SET)
+            lcmaps_log (LOG_WARNING, "%s: VERIFY_X509_CA_PATH already set...\n",logstr);
+        else if (setparam_result != VER_R_X509_PARAMS_OK)
         {
             verify_X509_term(&verify_data);
             goto fail_verify_proxy;
@@ -540,10 +613,10 @@ static int plugin_run_or_verify(
 
         if (pem_string)
         {
-            result = verify_X509_setParameter (&verify_data, VERIFY_X509_PRIVATEKEY_PEM, pem_string);
-            if (result == ERR_VERIFY_X509_PARAMS_ALREADY_SET)
-                Log (L_WARN,"VERIFY_X509_PRIVATEKEY_PEM already set...\n");
-            else if (result == ERR_VERIFY_X509_PARAMS_ACCESS_FAILURE)
+            setparam_result = verify_X509_setParameter (&verify_data, VERIFY_X509_PRIVATEKEY_PEM, pem_string);
+            if (setparam_result == VER_R_X509_PARAMS_ALREADY_SET)
+                lcmaps_log (LOG_WARNING,"%s: VERIFY_X509_PRIVATEKEY_PEM already set...\n",logstr);
+            else if (setparam_result != VER_R_X509_PARAMS_OK)
             {
                 verify_X509_term(&verify_data);
                 goto fail_verify_proxy;
@@ -553,8 +626,8 @@ static int plugin_run_or_verify(
         /* Allow for the ommission of a private key */
         if (!discard_private_key_absence)
         {
-            result = verify_X509_setParameter (&verify_data, VERIFY_X509_OPTIONS_MUST_HAVE_PRIV_KEY, VERIFY_ENABLE);
-            if (result == ERR_VERIFY_X509_PARAMS_ACCESS_FAILURE)
+            setparam_result = verify_X509_setParameter (&verify_data, VERIFY_X509_OPTIONS_MUST_HAVE_PRIV_KEY, VERIFY_ENABLE);
+            if (setparam_result != VER_R_X509_PARAMS_OK)
             {
                 verify_X509_term(&verify_data);
                 goto fail_verify_proxy;
@@ -563,10 +636,10 @@ static int plugin_run_or_verify(
 
         /* Require Limited proxies exclusively to succeed - Redundent / pedantic at this moment */
         if (allow_limited)
-            result = verify_X509_setParameter (&verify_data, VERIFY_X509_OPTIONS_ALLOW_LIMITED_PROXY, VERIFY_ENABLE);
+            setparam_result = verify_X509_setParameter (&verify_data, VERIFY_X509_OPTIONS_ALLOW_LIMITED_PROXY, VERIFY_ENABLE);
         else
-            result = verify_X509_setParameter (&verify_data, VERIFY_X509_OPTIONS_ALLOW_LIMITED_PROXY, VERIFY_DISABLE);
-        if (result == ERR_VERIFY_X509_PARAMS_ACCESS_FAILURE)
+            setparam_result = verify_X509_setParameter (&verify_data, VERIFY_X509_OPTIONS_ALLOW_LIMITED_PROXY, VERIFY_DISABLE);
+        if (setparam_result != VER_R_X509_PARAMS_OK)
         {
             verify_X509_term(&verify_data);
             goto fail_verify_proxy;
@@ -575,22 +648,19 @@ static int plugin_run_or_verify(
         /* Require Limited proxies exclusively to succeed - Redundent / pedantic at this moment */
         if (require_limited)
         {
-            result = verify_X509_setParameter (&verify_data, VERIFY_X509_OPTIONS_REQUIRE_LIMITED_PROXY, VERIFY_ENABLE);
-            if (result == ERR_VERIFY_X509_PARAMS_ACCESS_FAILURE)
+            setparam_result = verify_X509_setParameter (&verify_data, VERIFY_X509_OPTIONS_REQUIRE_LIMITED_PROXY, VERIFY_ENABLE);
+            if (setparam_result != VER_R_X509_PARAMS_OK)
             {
                 verify_X509_term(&verify_data);
                 goto fail_verify_proxy;
             }
         }
 
-        /* result = startVerifyProcess (fileName, CA_dir); */
+	/* returns 0 on success or ERR_get_error() value */
         result = verify_X509_verify(&verify_data);
         verify_X509_term(&verify_data);
-
-        if (result != X509_V_OK)
-        {
+        if (result)
             goto fail_verify_proxy;
-        }
     }
 
 
@@ -600,10 +670,11 @@ static int plugin_run_or_verify(
     {
         /* Select last certificate, check which type of certificate it is, and this MUST match with a limited proxy type */
         lastCert = sk_X509_value(myChain, 0);
-        certificate_type = lcmaps_type_of_proxy(lastCert);
-        if (!( (GT2_LIMITED_PROXY == certificate_type) || (RFC_LIMITED_PROXY == certificate_type) ))
-        {
-            lcmaps_log(LOG_NOTICE, "%s: The certificate chain does NOT contain have a Limited Proxy. This is REQUIRED by the configured policy.\n", logstr);
+        certificate_type = verify_type_of_proxy(lastCert);
+        if ( (certificate_type & LIMITED) != LIMITED) {
+            lcmaps_log(LOG_NOTICE,
+		"%s: The certificate chain does NOT contain a Limited Proxy."
+		" This is REQUIRED by the configured policy.\n", logstr);
             goto fail_verify_proxy;
         }
     }
@@ -616,7 +687,7 @@ static int plugin_run_or_verify(
     if (plt)
     {
         lcmaps_log_debug (1,"%s: Enforcing Proxy Life Time Policy\n", logstr);
-        if (!verifyProxyLifeTime (plt, myChain, depth))
+        if (!lcmaps_lifetime_verifyProxyLifeTime (plt, myChain, depth))
         {
             goto fail_verify_proxy;
         }
@@ -627,7 +698,7 @@ static int plugin_run_or_verify(
     if (vplt)
     {
         lcmaps_log_debug (1,"%s: Enforcing VOMS Life Time Policy.\n", logstr);
-        if (!verifyVOMSLifeTime (vplt, vomsdata))
+        if (!lcmaps_lifetime_verifyVOMSLifeTime (vplt, vomsdata))
         {
             goto fail_verify_proxy;
         }
@@ -645,31 +716,20 @@ fail_verify_proxy:
 
     lcmaps_log(LOG_INFO,"%s: verify proxy plugin failed\n", logstr);
 
+    while ((errcode=ERR_get_error())!=0)	{
+	reason=ERR_reason_error_string(errcode);
+	if (reason)
+	    lcmaps_log(LOG_DEBUG,"%s:  error queue: function %s (%s): %s\n",
+		    logstr,
+		    ERR_func_error_string(errcode),
+		    ERR_lib_error_string(errcode),
+		    reason);
+	else
+	    lcmaps_log(LOG_DEBUG,"%s:  error queue: %s\n",
+		    logstr,
+		    ERR_error_string(errcode,NULL));
+    }
+
     return LCMAPS_MOD_FAIL;
 }
 
-
-
-/******************************************************************************
-Function:   plugin_terminate
-Description:
-    Terminate plugin
-Parameters:
-
-Returns:
-    LCMAPS_MOD_SUCCESS : succes
-    LCMAPS_MOD_FAIL    : failure
-******************************************************************************/
-int plugin_terminate()
-{
-    char * logstr = "lcmaps_plugin_verify_proxy-plugin_terminate()";
-    lcmaps_log_debug(4,"%s: terminating\n", logstr);
-
-    /* if (vomsdir) free(vomsdir); */
-    if (certdir) free(certdir);
-
-    FreeProxyLevelTTL (&plt);
-    FreeProxyLevelTTL (&vplt);
-
-    return LCMAPS_MOD_SUCCESS;
-}
